from abc import ABC, abstractmethod


def to_snake_case(s):
    # TODO: implement and move to utils
    return s


class AbstractChannel(ABC):
    """Interface for channels."""

    @abstractmethod
    def put(self, context, name, value):
        pass

    @abstractmethod
    def get(self, context, name):
        pass


class InMemoryChannel(AbstractChannel):
    def __init__(self):
        self._stored = {}

    def put(self, task_name, value_name, value):
        if task_name in self._stored:
            pass


class TaskGraph:
    """
    Builds an abstract directed task graph.  The graph may then be instantiated
    as a concrete program in open-ended ways, e.g., as a single
    :class:`~AbstractTask` instance, or as an Airflow DAG.  Instantiation is
    essentially a process of interpretation of the graph in different contexts.

    The nodes of the graph are abstract tasks, preferably pure computations,
    while the arcs go from an output of one task to an input of another.  The
    effects that might be needed to do the actual transfer of data are in the
    interpretation of the arcs.  For example, storing and retrieving
    intermediate results on S3, in the case of interpretation in an Airflow DAG.
    """

    def __init__(self):
        self.nodes = {}
        self.arcs = {}

    def insert(self, task, task_name=None, inputs=None):
        """Inserts a task as a node and connects it to the outputs of an already
        inserted task.

        :param task: The task to insert.  Should be an instance of
            :class:`~AbstractTask`.
        :param task_name: The name of the node that represents the task.  Should
            not exist yet. Defaults to the snake_cased name of the class.
        :param inputs: A dictionary of inputs. Defaults to no inputs.

        >> graph = TaskGraph()
        >> outputs = graph.insert(FirstTask())
        >> graph.insert(SecondTask(), {"train_data": outputs["raw_data"]})

        :raises ValueError: If there is already a node with name `task_name` in
            the graph, or when the specified inputs do not match the expected
            inputs of the task.
        """
        self._check_inputs(task, inputs)
        node_name = task_name or to_snake_case(task.__class__.__name__)
        self.nodes[node_name] = task

        return self._create_outputs(task, node_name)

    def _check_inputs(self, task, inputs):
        # TODO: check whether the input specification (in terms of a
        # balsam.schema) matches the expected inputs for the task, and check
        # whether the producer already exists as a node name in the graph.
        pass

    def _create_outputs(self, task, node_name):
        return {
            name: {
                'schema': schema,
                'producer': node_name
            }
            for (name, schema) in task.output_schema.items()
        }


## === EXAMPLE CLIENT ====

# def create_graph():
#     graph = TaskGraph()
#     task1_outputs = graph.insert(Task1())
#     task2_outputs = graph.insert(Task2(),
#                                  {'bid_landscape': task2_outputs['bid_landscape']})



# pylint: disable=no-self-use
import json
import os
import uuid

import numpy as np
import pandas as pd
import pytest

from balsam.store import (ArrayRef, DataFrameRef, Encoding, FileRef, InMemoryRefStore,
                          JsonRef, LocalRefStore)
from test_balsam.utils import expand_path


@pytest.fixture
def store():
    return LocalRefStore()


@pytest.fixture
def mock_store():
    return InMemoryRefStore()


class TestLocalRefStore:
    def test_reading_data_frames(self, store):
        ref = DataFrameRef(expand_path('df.csv'))
        df = ref.parse(store)
        assert isinstance(df, pd.DataFrame)

    def test_write_binary_data(self, store):
        path = f'{uuid.uuid4()}.bin'
        try:
            ref = FileRef(path, encoding=Encoding.BINARY)
            ref.write(b'xxx', store)
            value = ref.read(store)
            assert value == b'xxx'
        finally:
            os.remove(path)

    def test_write_string_data(self, store):
        path = f'{uuid.uuid4()}.bin'
        try:
            ref = FileRef(path, encoding=Encoding.STRING)
            ref.write('xxx', store)
            value = ref.read(store)
            assert value == 'xxx'
        finally:
            os.remove(path)


class TestDataFrameRef:
    def test_write_data_frame_to_buffer(self, store):
        df = pd.DataFrame([1])
        path = expand_path('not_used.csv')
        ref = DataFrameRef(path)
        ref.write(df, store)

        actual = pd.read_csv(path, index_col=0)
        os.remove(path)

        assert (actual.values == df.values).all()

    def test_read_csv_data_frame_ref(self, mock_store):
        ref = DataFrameRef('df.csv')
        expected = pd.DataFrame({'x': [1], 'y': [2]})
        ref.write(expected, mock_store)

        actual = ref.read(mock_store)
        assert actual.equals(expected)

    def test_read_tsv_data_frame_ref(self, mock_store):
        mock_store.write_raw('df.tsv', 'x\ty\n1\t2')
        ref = DataFrameRef('df.tsv')
        expected = pd.DataFrame({'x': [1], 'y': [2]})

        actual = ref.read(mock_store)
        assert actual.equals(expected)


class TestJsonRef:
    def test_storing_json_values(self, mock_store):
        ref = JsonRef('single.json')
        ref.write({'x': 1}, mock_store)
        assert ref.read(mock_store) == {'x': 1}

    def test_storing_json_values_with_data_frames(self, mock_store):
        ref = JsonRef('compound.json')
        df1 = pd.DataFrame([1])
        df2 = pd.DataFrame([2])
        ref.write({'df1': df1, 'df2': df2}, mock_store)

        assert (DataFrameRef('df1.csv').read(mock_store).values == df1.values).all()
        assert (DataFrameRef('df2.csv').read(mock_store).values == df2.values).all()
        assert json.loads(mock_store.read_raw('compound.json')) == {
            'df1': 'df1.csv',
            'df2': 'df2.csv'
        }

    def test_storing_json_values_with_arrays(self, mock_store):
        ref = JsonRef('compound.json')
        arr1 = np.array([1, 2])
        arr2 = np.array([3, 4])
        ref.write({'arr1': arr1, 'arr2': arr2}, mock_store)

        assert (ArrayRef('arr1.csv').read(mock_store) == arr1).all()
        assert (ArrayRef('arr2.csv').read(mock_store) == arr2).all()
        assert json.loads(mock_store.read_raw('compound.json')) == {
            'arr1': 'arr1.csv',
            'arr2': 'arr2.csv'
        }

    def test_store_values_references_in_json_in_same_dir(self, mock_store):
        ref = JsonRef('dir/compound.json')
        arr1 = np.array([1, 2])
        ref.write({'arr1': arr1}, mock_store)

        assert (ArrayRef('dir/arr1.csv').read(mock_store) == arr1).all()

    def test_reading_json_values(self, mock_store):
        ref = JsonRef('single.json')
        ref.write({'x': 1}, mock_store)
        assert ref.read(mock_store) == {'x': 1}

    def test_reading_json_values_with_references(self, mock_store):
        ref = JsonRef('compound.json')
        ref.write({'arr1': 'array1.csv', 'arr2': 'array2.csv'}, mock_store)
        arr1 = np.array([1, 2])
        arr2 = np.array([3, 4])
        arr1_ref = ArrayRef('array1.csv')
        arr2_ref = ArrayRef('array2.csv')
        arr1_ref.write(arr1, mock_store)
        arr2_ref.write(arr2, mock_store)

        actual = ref.read(mock_store)
        assert set(actual.keys()) == {'arr1', 'arr2'}
        assert (actual['arr1'] == arr1).all()
        assert (actual['arr2'] == arr2).all()



import cerberus


class ValidationError(Exception):
    """Exception thrown when validation fails."""

    def __init__(self, document, errors, info=None):
        super().__init__(self, "Failed to validate")
        self.errors = errors
        self.info = info
        self._document = document

    def __repr__(self):
        if self.info:
            return "<{}({}, {}, {})>".format(self.__class__.__name__, self._document,
                                             self.errors, self.info)
        return "<{}({}, {})>".format(self.__class__.__name__, self._document,
                                     self.errors)


def check(schema, document, info=None):
    """Validates whether the document complies with the schema.  Aims to
    abstract away some details of the schema library from the client.

    :param schema: A schema constructed from balsam types.
    :param document: The value to check against the schema.

    :raises ValidationError: When the document does not comply with the schema.

    :returns: The document.
    """
    v = cerberus.Validator(schema, allow_unknown=True)
    if not v.validate(document):
        raise ValidationError(document, v.errors, info)
    return document


def integer(required=True):
    return {'required': required, 'type': 'integer'}


def list_of(schema, required=True):
    nullable = not required
    return {'type': 'list', 'required': required, 'nullable': nullable, 'schema': schema}


def data_frame():
    return {'required': True, 'type': 'data_frame'}


def date():
    return {'required': True, 'type': 'date'}


def string(required=True):
    return {'required': required, 'type': 'string'}


def floating():
    return {'required': True, 'type': 'float'}


def enum(values):
    return {'required': True, 'type': 'string', 'allowed': values}


def boolean():
    return {'type': 'boolean'}


def dict_of(schema, required=True):
    nullable = not required
    return {'type': 'dict', 'required': required, 'nullable': nullable, 'schema': schema}


def array():
    return {'required': True, 'type': 'array'}


def datetime(required=True):
    nullable = not required
    return {'required': required, 'nullable': nullable, 'type': 'datetime'}


def selection(required=True):
    nullable = not required
    return {'required': required, 'nullable': nullable, 'type': 'selection'}



import numpy as np
import pandas as pd
import pytest

from balsam.schema import (ValidationError, array, check, data_frame, enum, integer,
                           list_of, dict_of, selection)
from balsam.selection import Contains, And, In


class BaseTestValidate:
    schema = None
    valid = None
    invalid = None

    def test_invalid_value_raises_error(self):
        with pytest.raises(ValidationError):
            check({"value": self.schema}, {"value": self.invalid})

    def test_valid_value_passes_check(self):
        try:
            check({"value": self.schema}, {"value": self.valid})
        except ValidationError as exc:
            assert False, exc


class TestValidateDataFrame(BaseTestValidate):
    schema = data_frame()
    valid = pd.DataFrame([1])
    invalid = []


class TestValidateInteger(BaseTestValidate):
    schema = integer()
    valid = 1
    invalid = "string"


class TestValidateEnum(BaseTestValidate):
    schema = enum(["x", "y"])
    valid = "x"
    invalid = "z"


class TestValidateArray(BaseTestValidate):
    schema = array()
    valid = np.array([1])
    invalid = [1]


class TestNonRequiredList(BaseTestValidate):
    schema = dict_of({'x': list_of(integer(), required=False)})
    valid = {}
    invalid = {'x': [1.2]}


class TestValidateSelection(BaseTestValidate):
    schema = selection()
    valid = And(In('campaignId', [1, 2, 3]), Contains('campaignText', 'nonBranded'))
    invalid = 'hello'



import inspect
from balsam.selection import base


def from_dict(obj):
    """Convert Python native object to a selection.

    :param obj: The object to convert
    :rtype: A selection, i.e., an instance of a subclass of :class:`~Selection`.
    """

    def _from_dict(_obj):
        if 'operation' in _obj:
            # Deserializes to a subclass of BinaryOperator
            for Op in _binary_operations():
                if Op.op == _obj['operation']:
                    try:
                        return Op(_from_dict(_obj['left']), _from_dict(_obj['right']))
                    except KeyError as exc:
                        raise ValueError("Could not convert {} to selection: {}".format(
                            obj, exc))
        if 'sql' in _obj:
            return base.SQL(_obj['sql'])
        return _obj

    conversion = _from_dict(obj)
    if not isinstance(conversion, base.Selection):
        raise ValueError("Could not convert {} to selection".format(obj))
    return conversion


def _binary_operations():
    for name in dir(base):
        item = getattr(base, name)
        if inspect.isclass(item) and issubclass(item, base.BinaryOperator):
            yield getattr(base, name)
        yield base.In



import datetime
import json

from balsam import selection


class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        # pylint: disable=arguments-differ, method-hidden
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        if isinstance(obj, datetime.date):
            return obj.isoformat()
        if isinstance(obj, datetime.timedelta):
            return (datetime.datetime.min + obj).time().isoformat()
        if isinstance(obj, selection.Selection):
            return selection.to_dict(obj)

        return super(CustomEncoder, self).default(obj)



import datetime as dt

import pandas as pd
import pytest

from balsam.schema import (ValidationError, boolean, data_frame, date, dict_of, enum,
                           floating, integer, list_of)
from balsam.store import DataFrameRef, InMemoryRefStore, JsonRef, LocalRefStore
from balsam.tasks import AbstractTask, create_argument_parser, run_as_cli_program
from test_balsam.utils import expand_path


@pytest.fixture
def mock_store():
    return InMemoryRefStore()


@pytest.fixture
def ref_store():
    return LocalRefStore()


@pytest.fixture
def df():
    """Return the content of the `df.csv` fixture as a data frame."""
    return pd.DataFrame({1: [3], 2: [4]})


class TestRunAsCliProgram:
    def test_inputs_are_passed_to_task(self):
        task = _create_test_task({'initial': integer()}, {})
        output = run_as_cli_program(task, '--initial 1'.split())
        assert output == {'input/initial': 1}

    def test_settings_are_passed_to_task(self):
        task = _create_test_task({}, {'iterations': integer()})
        output = run_as_cli_program(task, '--iterations 100'.split())
        assert output == {'settings/iterations': 100}

    def test_does_not_load_data_frame_without_store(self):
        task = _create_test_task({'data': data_frame()}, {})
        with pytest.raises(ValidationError):
            run_as_cli_program(task, ['--data', expand_path('df.csv')])

    def test_load_data_frame_with_store(self, mock_store):
        task = _create_test_task({'data': data_frame()}, {})
        df = pd.DataFrame([1])
        DataFrameRef('df.csv').write(df, mock_store)

        output = run_as_cli_program(task, '--data df.csv'.split(), mock_store)
        assert (output['input/data'].values == df.values).all()

    def test_input_parameters_are_required_by_default(self, mock_store):
        task = _create_test_task({'data': data_frame()}, {})
        with pytest.raises(SystemExit):
            run_as_cli_program(task, [], mock_store)

    def test_settings_parameters_are_required_by_default(self, mock_store):
        task = _create_test_task({}, {'iterations': integer()})
        with pytest.raises(SystemExit):
            run_as_cli_program(task, [], mock_store)

    def test_arbitrary_output_params_are_not_required(self, mock_store):
        task = _create_test_task({}, {}, {'out': integer()})
        try:
            run_as_cli_program(task, [], mock_store)
        except SystemExit as exc:
            assert False, exc

    def test_data_frame_output_params_values_have_defaults(self, mock_store):
        task = _create_test_task({'df': data_frame()}, {}, {'input/df': data_frame()})
        expected = pd.DataFrame([1])
        DataFrameRef('in.csv').write(expected, mock_store)

        run_as_cli_program(task, '--df in.csv'.split(), mock_store)

        actual = DataFrameRef('input/df.csv').parse(mock_store)
        assert actual.values == expected.values

    def test_data_frame_outputs_are_written_to_store(self, mock_store):
        task = _create_test_task({'df': data_frame()}, {}, {'input/df': data_frame()})
        expected = pd.DataFrame([1])
        DataFrameRef('in.csv').write(expected, mock_store)

        run_as_cli_program(task, '--df in.csv --input/df out.csv'.split(), mock_store)

        actual = DataFrameRef('out.csv').parse(mock_store)
        assert actual.values == expected.values

    def test_json_outputs_are_written_to_store(self, mock_store):
        task = _create_output_task({
            'result': dict_of({
                'count': integer()
            })
        }, {'result': {
            'count': 1
        }})
        run_as_cli_program(task, '--result file.json'.split(), mock_store)

        assert JsonRef('file.json').parse(mock_store) == {'count': 1}

    def test_json_inputs_are_read_from_store(self, mock_store):
        value = {'count': 1}
        task = _create_test_task({'data': dict_of({'count': integer()})}, {})
        JsonRef('file.json').write(value, mock_store)
        output = run_as_cli_program(task, '--data file.json'.split(), mock_store)
        assert output['input/data'] == value


class TestCreateCommandLineParser:
    def test_parse_integers(self):
        task = _create_test_task({'initial': integer()}, {'count': integer()})
        parser = create_argument_parser(task)
        values = parser.parse_args(['--initial', '0', '--count', '1'])
        assert values.initial == 0
        assert values.count == 1

    def test_parse_list_of_integers(self):
        task = _create_test_task({'initial': list_of(integer())}, {'count': integer()})
        parser = create_argument_parser(task)
        values = parser.parse_args('--initial 0 1 2 --count 1'.split())
        assert values.initial == [0, 1, 2]
        assert values.count == 1

    def test_parse_data_frame_reference(self):
        task = _create_test_task({}, {'df': data_frame()})
        parser = create_argument_parser(task)
        values = parser.parse_args('--df data.csv'.split())
        assert values.df == DataFrameRef('data.csv')

    def test_parse_floating_point(self):
        task = _create_test_task({}, {'real': floating()})
        parser = create_argument_parser(task)
        values = parser.parse_args('--real 0.9'.split())
        assert values.real == 0.9

    def test_parse_enum(self):
        task = _create_test_task({}, {'value': enum(["x", "y"])})
        parser = create_argument_parser(task)
        values = parser.parse_args('--value x'.split())
        assert values.value == "x"

    def test_parse_dict_as_json_ref(self):
        task = _create_test_task({}, {'json': dict_of({'n': integer()})}, {})
        parser = create_argument_parser(task)
        values = parser.parse_args('--json file.json'.split())
        assert values.json == JsonRef('file.json')

    def test_use_default_value_when_json_ref_is_missing(self):
        task = task = _create_test_task({}, {}, {'results': dict_of({'n': integer()})})
        parser = create_argument_parser(task)
        values = parser.parse_args([])
        assert values.results == JsonRef('results.json')

    def test_parse_enum_fails_when_argument_not_allowed(self):
        task = _create_test_task({}, {'value': enum(["x", "y"])})
        parser = create_argument_parser(task)
        with pytest.raises(SystemExit):
            parser.parse_args('--value z'.split())

    def test_parse_presence_of_flag_as_true(self):
        task = _create_test_task({}, {'flag': boolean()})
        parser = create_argument_parser(task)
        value = parser.parse_args(['--flag'])
        assert value.flag is True

    def test_parse_absence_of_flag_as_false(self):
        task = _create_test_task({}, {'flag': boolean()})
        parser = create_argument_parser(task)
        value = parser.parse_args([])
        assert value.flag is False

    def test_parse_date(self):
        task = _create_test_task({}, {'date': date()})
        parser = create_argument_parser(task)
        value = parser.parse_args('--date 2019-01-01'.split())
        assert value.date == dt.date(2019, 1, 1)


def _create_test_task(input_schema, settings_schema, output_schema=None):
    class TestTask(AbstractTask):
        @property
        def input_schema(self):
            return input_schema

        @property
        def output_schema(self):
            return output_schema or {}

        @property
        def settings_schema(self):
            return settings_schema

        def run(self, settings, inputs):
            result = {}
            for (key, value) in settings.items():
                result['settings/{}'.format(key)] = value
            for (key, value) in inputs.items():
                result['input/{}'.format(key)] = value
            return result

    return TestTask()


def _create_output_task(output_schema, output):
    class OutputTestTask(AbstractTask):
        @property
        def input_schema(self):
            return {}

        @property
        def output_schema(self):
            return output_schema

        @property
        def settings_schema(self):
            return {}

        def run(self, _settings, _inputs):
            return output

    return OutputTestTask()
